#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""NiBabies runner."""
from .. import config


def main():
    """Entry point."""
    import gc
    import sys
    from multiprocessing import Manager, Process
    from os import EX_SOFTWARE
    from pathlib import Path

    from ..utils.bids import write_bidsignore, write_derivative_description
    from .parser import parse_args

    parse_args()

    # sentry_sdk = None
    # if not config.execution.notrack:
    #     import sentry_sdk
    #     from ..utils.sentry import sentry_setup

    #     sentry_setup()

    # CRITICAL Save the config to a file. This is necessary because the execution graph
    # is built as a separate process to keep the memory footprint low. The most
    # straightforward way to communicate with the child process is via the filesystem.
    config_file = config.execution.work_dir / config.execution.run_uuid / "config.toml"
    config_file.parent.mkdir(exist_ok=True, parents=True)
    config.to_filename(config_file)

    # CRITICAL Call build_workflow(config_file, retval) in a subprocess.
    # Because Python on Linux does not ever free virtual memory (VM), running the
    # workflow construction jailed within a process preempts excessive VM buildup.
    with Manager() as mgr:
        from .workflow import build_workflow

        retval = mgr.dict()
        p = Process(target=build_workflow, args=(str(config_file), retval))
        p.start()
        p.join()

        retcode = p.exitcode or retval.get("return_code", 0)
        nibabies_wf = retval.get("workflow", None)

    # CRITICAL Load the config from the file. This is necessary because the ``build_workflow``
    # function executed constrained in a process may change the config (and thus the global
    # state of NiBabies).
    config.load(config_file)

    if config.execution.reports_only:
        sys.exit(int(retcode > 0))

    if nibabies_wf and config.execution.write_graph:
        nibabies_wf.write_graph(graph2use="colored", format="svg", simple_form=True)

    retcode = retcode or (nibabies_wf is None) * EX_SOFTWARE
    if retcode != 0:
        sys.exit(retcode)

    # Generate boilerplate
    with Manager() as mgr:
        from .workflow import build_boilerplate

        p = Process(target=build_boilerplate, args=(str(config_file), nibabies_wf))
        p.start()
        p.join()

    if config.execution.boilerplate_only:
        sys.exit(int(retcode > 0))

    # Clean up master process before running workflow, which may create forks
    gc.collect()

    # Sentry tracking
    # if sentry_sdk is not None:
    #     with sentry_sdk.configure_scope() as scope:
    #         scope.set_tag("run_uuid", config.execution.run_uuid)
    #         scope.set_tag("npart", len(config.execution.participant_label))
    #     sentry_sdk.add_breadcrumb(message="nibabies started", level="info")
    #     sentry_sdk.capture_message("nibabies started", level="info")

    config.loggers.workflow.log(
        15,
        "\n".join(["nibabies config:"] + ["\t\t%s" % s for s in config.dumps().splitlines()]),
    )
    config.loggers.workflow.log(25, "nibabies started!")
    # errno = 1  # Default is error exit unless otherwise set
    try:
        nibabies_wf.run(**config.nipype.get_plugin())
    except Exception as e:
        # if not config.execution.notrack:
        #     from ..utils.sentry import process_crashfile

        #     crashfolders = [
        #         config.execution.nibabies_dir,
        #         / "sub-{}".format(s)
        #         / "log"
        #         / config.execution.run_uuid
        #         for s in config.execution.participant_label
        #     ]
        #     for crashfolder in crashfolders:
        #         for crashfile in crashfolder.glob("crash*.*"):
        #             process_crashfile(crashfile)

        #     if "Workflow did not execute cleanly" not in str(e):
        #         sentry_sdk.capture_exception(e)
        config.loggers.workflow.critical("nibabies failed: %s", e)
        raise
    else:
        config.loggers.workflow.log(25, "nibabies finished successfully!")
        # if not config.execution.notrack:
        #     success_message = "nibabies finished without errors"
        #     sentry_sdk.add_breadcrumb(message=success_message, level="info")
        #     sentry_sdk.capture_message(success_message, level="info")

        # Bother users with the boilerplate only iff the workflow went okay.
        boiler_file = config.execution.nibabies_dir / "logs" / "CITATION.md"
        if boiler_file.exists():
            if config.environment.exec_env in (
                "singularity",
                "docker",
                "nibabies-docker",
            ):
                boiler_file = Path("<OUTPUT_PATH>") / boiler_file.relative_to(
                    config.execution.output_dir
                )
            config.loggers.workflow.log(
                25,
                "Works derived from this nibabies execution should include the "
                f"boilerplate text found in {boiler_file}.",
            )

        if config.workflow.run_reconall:
            from niworkflows.utils.misc import _copy_any
            from templateflow import api

            dseg_tsv = str(api.get("fsaverage", suffix="dseg", extension=[".tsv"]))
            _copy_any(dseg_tsv, str(config.execution.nibabies_dir / "desc-aseg_dseg.tsv"))
            _copy_any(dseg_tsv, str(config.execution.nibabies_dir / "desc-aparcaseg_dseg.tsv"))
        # errno = 0
    finally:
        from pkg_resources import resource_filename as pkgrf

        from ..reports.core import generate_reports

        # Generate reports phase
        generate_reports(
            config.execution.participant_label,
            config.execution.nibabies_dir,
            config.execution.run_uuid,
            config=pkgrf("nibabies", "data/reports-spec.yml"),
            packagename="nibabies",
        )
        write_derivative_description(config.execution.bids_dir, config.execution.nibabies_dir)
        write_bidsignore(config.execution.nibabies_dir)

        # if failed_reports and not config.execution.notrack:
        #     sentry_sdk.capture_message(
        #         "Report generation failed for %d subjects" % failed_reports,
        #         level="error",
        #     )
        # sys.exit(int((errno + failed_reports) > 0))


if __name__ == "__main__":
    raise RuntimeError(
        "Please `pip install` this and run via the commandline interfaces, `nibabies <command>`"
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Multi-echo EPI
~~~~~~~~~~~~~~

For using multi-echo EPI data.

Change directory to provide relative paths for doctests
>>> import os
>>> filepath = os.path.dirname( os.path.realpath( __file__ ) )
>>> datadir = os.path.realpath(os.path.join(filepath, '../data/'))
>>> os.chdir(datadir)

"""
import os

from nipype import logging
from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    File,
    TraitedSpec,
    traits,
)

LOGGER = logging.getLogger("nipype.interface")


class T2SMapInputSpec(CommandLineInputSpec):
    in_files = traits.List(
        File(exists=True),
        argstr="-d %s",
        position=1,
        mandatory=True,
        minlen=3,
        desc="multi-echo BOLD EPIs",
    )
    echo_times = traits.List(
        traits.Float, argstr="-e %s", position=2, mandatory=True, minlen=3, desc="echo times"
    )
    fittype = traits.Enum(
        "curvefit",
        "loglin",
        argstr="--fittype %s",
        position=3,
        usedefault=True,
        desc=(
            "Desired fitting method: "
            '"loglin" means that a linear model is fit '
            "to the log of the data. "
            '"curvefit" means that a more computationally '
            "demanding monoexponential model is fit "
            "to the raw data."
        ),
    )


class T2SMapOutputSpec(TraitedSpec):
    t2star_map = File(exists=True, desc="limited T2* map")
    s0_map = File(exists=True, desc="limited S0 map")
    optimal_comb = File(exists=True, desc="optimally combined ME-EPI time series")


class T2SMap(CommandLine):
    """
    Runs the tedana T2* workflow to generate an adaptive T2* map and create
    an optimally combined ME-EPI time series.

    Example
    =======
    >>> from nibabies.interfaces import multiecho
    >>> t2smap = multiecho.T2SMap()
    >>> t2smap.inputs.in_files = [
    ...     data_dir / 'sub-01_run-01_echo-1_bold.nii.gz',
    ...     data_dir / 'sub-01_run-01_echo-2_bold.nii.gz',
    ...     data_dir / 'sub-01_run-01_echo-3_bold.nii.gz']
    >>> t2smap.inputs.echo_times = [0.013, 0.027, 0.043]
    >>> t2smap.cmdline
    't2smap -d .../sub-01_run-01_echo-1_bold.nii.gz \
    .../sub-01_run-01_echo-2_bold.nii.gz \
    .../sub-01_run-01_echo-3_bold.nii.gz -e 13.0 27.0 43.0 --fittype curvefit'
    """

    _cmd = "t2smap"
    input_spec = T2SMapInputSpec
    output_spec = T2SMapOutputSpec

    def _format_arg(self, name, trait_spec, value):
        if name == "echo_times":
            value = [te * 1000 for te in value]
        return super(T2SMap, self)._format_arg(name, trait_spec, value)

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_dir = os.getcwd()
        outputs["t2star_map"] = os.path.join(out_dir, "T2starmap.nii.gz")
        outputs["s0_map"] = os.path.join(out_dir, "S0map.nii.gz")
        outputs["optimal_comb"] = os.path.join(out_dir, "desc-optcom_bold.nii.gz")
        return outputs



# Use infant_recon_all to generate subcortical segmentations and cortical parcellations

from nipype.interfaces import freesurfer as fs
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.interfaces.freesurfer import PatchedLTAConvert as LTAConvert
from smriprep.workflows.surfaces import init_gifti_surface_wf

from ...interfaces.freesurfer import InfantReconAll


def init_infant_surface_recon_wf(*, age_months, use_aseg=False, name="infant_surface_recon_wf"):
    from niworkflows.engine.workflows import LiterateWorkflow

    wf = LiterateWorkflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                "subjects_dir",
                "subject_id",
                "anat_orig",
                "anat_skullstripped",
                "anat_preproc",
                "anat_aseg",
                "t2w",
            ],
        ),
        name="inputnode",
    )
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                "subjects_dir",
                "subject_id",
                "anat2fsnative_xfm",
                "fsnative2anat_xfm",
                "surfaces",
                "anat_aseg",
                "anat_aparc",
            ]
        ),
        name="outputnode",
    )

    wf.__desc__ = f"""\
Brain surfaces were reconstructed using `infant_recon_all` [FreeSurfer
{fs.Info().looseversion() or "<ver>"}, RRID:SCR_001847, @infantfs],
leveraging the masked, preprocessed T1w and anatomical segmentation.
"""

    gen_recon_outdir = pe.Node(niu.Function(function=_gen_recon_dir), name="gen_recon_outdir")

    # inject the intensity-normalized skull-stripped t1w from the brain extraction workflow
    recon = pe.Node(InfantReconAll(age=age_months), name="reconall")

    # these files are created by babyFS, but transforms are for masked anatomicals
    # https://github.com/freesurfer/freesurfer/blob/
    # 8b40551f096294cc6603ce928317b8df70bce23e/infant/infant_recon_all#L744
    # TODO: calculate full anat -> fsnative transform?
    get_tal_lta = pe.Node(
        niu.Function(function=_get_talairch_lta),
        name="get_tal_xfm",
    )
    fsnative2anat_xfm = pe.Node(
        LTAConvert(out_lta=True, invert=True),
        name="fsnative2anat_xfm",
    )

    # convert generated surfaces to GIFTIs
    gifti_surface_wf = init_gifti_surface_wf()

    get_aseg = pe.Node(niu.Function(function=_get_aseg), name="get_aseg")
    get_aparc = pe.Node(niu.Function(function=_get_aparc), name="get_aparc")
    aparc2nii = pe.Node(fs.MRIConvert(out_type="niigz"), name="aparc2nii")

    if use_aseg:
        # TODO: Add precomputed segmentation upon new babyFS rel
        wf.connect(inputnode, "anat_aseg", recon, "aseg_file")

    # fmt: off
    wf.connect([
        (inputnode, gen_recon_outdir, [
            ('subjects_dir', 'subjects_dir'),
            ('subject_id', 'subject_id'),
        ]),
        (inputnode, recon, [
            ('anat_skullstripped', 'mask_file'),
            ('subject_id', 'subject_id'),
        ]),
        (gen_recon_outdir, recon, [
            ('out', 'outdir'),
        ]),
        (recon, outputnode, [
            ('subject_id', 'subject_id'),
            (('outdir', _parent), 'subjects_dir'),
        ]),
        (recon, gifti_surface_wf, [
            ('subject_id', 'inputnode.subject_id'),
            (('outdir', _parent), 'inputnode.subjects_dir'),
        ]),
        (recon, get_aparc, [
            ('outdir', 'fs_subject_dir'),
        ]),
        (recon, get_aseg, [
            ('outdir', 'fs_subject_dir'),
        ]),
        (get_aseg, outputnode, [
            ('out', 'anat_aseg'),
        ]),
        (get_aparc, aparc2nii, [
            ('out', 'in_file'),
        ]),
        (aparc2nii, outputnode, [
            ('out_file', 'anat_aparc'),
        ]),
        (recon, get_tal_lta, [
            ('outdir', 'fs_subject_dir'),
        ]),
        (get_tal_lta, outputnode, [
            ('out', 'anat2fsnative_xfm'),
        ]),
        (get_tal_lta, fsnative2anat_xfm, [
            ('out', 'in_lta'),
        ]),
        (fsnative2anat_xfm, outputnode, [
            ('out_lta', 'fsnative2anat_xfm'),
        ]),
        (fsnative2anat_xfm, gifti_surface_wf, [
            ('out_lta', 'inputnode.fsnative2t1w_xfm')]),
        (gifti_surface_wf, outputnode, [
            ('outputnode.surfaces', 'surfaces'),
        ]),
    ])
    # fmt: on
    return wf


def _parent(p):
    from pathlib import Path

    return str(Path(p).parent)


def _gen_recon_dir(subjects_dir, subject_id):
    from pathlib import Path

    p = Path(subjects_dir) / subject_id
    p.mkdir(parents=True, exist_ok=True)
    return str(p)


def _get_talairch_lta(fs_subject_dir):
    """Fetch pre-computed transform from infant_recon_all"""
    from pathlib import Path

    xfm = Path(fs_subject_dir) / "mri" / "transforms" / "niftyreg_affine.lta"
    if not xfm.exists():
        raise FileNotFoundError("Could not find talairach transform.")
    return str(xfm.absolute())


def _get_aseg(fs_subject_dir):
    """Fetch infant_recon_all's aparc+aseg"""
    from pathlib import Path

    aseg = Path(fs_subject_dir) / "mri" / "aseg.nii.gz"
    if not aseg.exists():
        raise FileNotFoundError("Could not find aseg.")
    return str(aseg)


def _get_aparc(fs_subject_dir):
    """Fetch infant_recon_all's aparc+aseg"""
    from pathlib import Path

    aparc = Path(fs_subject_dir) / "mri" / "aparc+aseg.mgz"
    if not aparc.exists():
        raise FileNotFoundError("Could not find aparc.")
    return str(aparc)
